"""ThermoWorks Cloud client library.

This module provides the primary interface for interacting with the ThermoWorks Cloud
service, allowing users to access their ThermoWorks Cloud devices and data.

This is the main public module of the thermoworks_cloud package.
"""

import logging

from . import models
from .auth import Auth, AuthenticationError, AuthenticationErrorReason, AuthFactory
from .core import ThermoworksCloud, ResourceNotFoundError

# The publicly accessible classes for this module
__all__ = [
    "ThermoworksCloud",
    "Auth",
    "AuthFactory",
    "AuthenticationError",
    "AuthenticationErrorReason",
    "ResourceNotFoundError",
    "models",
]

# Tells pdoc how to parse the doc strings in this module
__docformat__ = "google"

# Allows for consuming applications to look at log messages if they'd like
_LOGGER = logging.getLogger(__name__)
_LOGGER.addHandler(logging.NullHandler())



"""Classes related to a Device."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Dict

from thermoworks_cloud.utils import parse_datetime, map_firestore_fields, parse_nested_object


@dataclass
class BigQueryInfo:
    """BigQueryInfo contains information about the BigQuery table for a device."""

    table_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    dataset_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})


@dataclass
class Device:  # pylint: disable=too-many-instance-attributes
    """Device contains information about a Thermoworks device.

    All fields are optional as different device types may have different properties.
    The Node device type is known to have all these properties.
    """

    device_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    serial: Optional[str] = field(default=None, metadata={
                                  "firestore_type": "stringValue"})
    label: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    type: Optional[str] = field(default=None, metadata={
                                "firestore_type": "stringValue"})
    firmware: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    color: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    thumbnail: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    device_display_units: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    iot_device_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    device_name: Optional[str] = field(
        default=None, metadata={"api_name": "device", "firestore_type": "stringValue"})
    account_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    status: Optional[str] = field(default=None, metadata={
                                  "firestore_type": "stringValue"})
    battery_state: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    big_query_info: Optional[BigQueryInfo] = field(
        default=None, metadata={"api_name": "bigQuery"})
    battery: Optional[int] = field(
        default=None,
        metadata={
            "firestore_type": "integerValue",
            "converter": int
        }
    )
    wifi_strength: Optional[int] = field(
        default=None,
        metadata={
            "api_name": "wifi_stength",
            "firestore_type": "integerValue", "converter": int
        }
    )
    recording_interval_in_seconds: Optional[int] = field(
        default=None, metadata={"firestore_type": "integerValue", "converter": int})
    transmit_interval_in_seconds: Optional[int] = field(
        default=None, metadata={"firestore_type": "integerValue", "converter": int})
    pending_load: Optional[bool] = field(
        default=None, metadata={"firestore_type": "booleanValue"})
    battery_alert_sent: Optional[bool] = field(
        default=None, metadata={"firestore_type": "booleanValue"})
    export_version: Optional[float] = field(
        default=None, metadata={"firestore_type": "doubleValue"})
    last_seen: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    last_purged: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    last_archive: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    last_telemetry_saved: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    last_wifi_connection: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    last_bluetooth_connection: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    session_start: Optional[datetime] = field(
        default=None,
        metadata={"firestore_type": "timestampValue", "converter": parse_datetime})
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None

    # Dictionary to store any additional properties not explicitly defined
    additional_properties: Optional[Dict] = None


def _document_to_device(document: dict) -> Device:
    """Convert a Firestore Document object into a Device object."""
    fields = document["fields"]
    device = map_firestore_fields(fields, Device)

    # Handle BigQuery info separately since it's a nested object
    if "bigQuery" in fields and "mapValue" in fields["bigQuery"]:
        try:
            device.big_query_info = parse_nested_object(
                fields["bigQuery"]["mapValue"], BigQueryInfo)
        except (KeyError, TypeError):
            device.big_query_info = None

    # Document timestamps
    if "createTime" in document:
        device.create_time = parse_datetime(document["createTime"])
    if "updateTime" in document:
        device.update_time = parse_datetime(document["updateTime"])

    return device



"""Configures pytest for all tests, providing various fixtures"""

from unittest.mock import patch
from aiohttp import ClientSession
import pytest
from pytest_httpserver import HTTPServer

from tests.core_test_object import CoreTestObject
from tests.test_data import CONFIG_RETURN_VALUE
from tests.test_data import LOGIN_RETURN_VALUE, TEST_EMAIL_ADDRESS, TEST_PASSWORD
from tests.auth_test_object import AuthTestObject
from thermoworks_cloud.auth import _TokenManager, Auth, AuthFactory


@pytest.fixture(autouse=True, name="httpserver_endpoint")
def fixture_httpserver_endpoint(httpserver: HTTPServer) -> str:
    """Provides a fixture that returns the endpoint of the httpserver."""
    return f"http://{httpserver.host}:{httpserver.port}"


@pytest.fixture(autouse=True)
def override_token_manager_hosts(httpserver_endpoint: str):
    """Override the hosts that _TokenManager uses so that it will call our mock http server"""

    with patch.object(_TokenManager, "_IDENTITY_HOST", httpserver_endpoint):
        with patch.object(_TokenManager, "_TOKEN_HOST", httpserver_endpoint):
            yield


@pytest.fixture(autouse=True)
def override_authfactory_hosts(httpserver_endpoint: str):
    """Override the hosts that AuthFactory uses so that it will call our mock http server"""

    with patch.object(AuthFactory, "_FIREBASE_HOST", httpserver_endpoint):
        with patch.object(AuthFactory, "_FIRESTORE_HOST", httpserver_endpoint):
            yield


@pytest.fixture(name="auth_test_object")
def fixture_auth_test_object(httpserver: HTTPServer) -> AuthTestObject:
    """Provide an AuthTestObject which uses the mock http server."""
    return AuthTestObject(httpserver)


@pytest.fixture(name="client_session")
async def fixture_client_session():
    """Provide a ClientSession for use in tests."""
    session = ClientSession()
    yield session

    # Runs after the test completes
    await session.close()


@pytest.fixture(name="auth")
async def fixture_auth(
    client_session: ClientSession, auth_test_object: AuthTestObject
) -> Auth:
    """Provides an Auth object for use in component tests."""
    # Setup mock responses
    auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
    auth_test_object.expect_login(TEST_EMAIL_ADDRESS, TEST_PASSWORD).respond_with_json(
        LOGIN_RETURN_VALUE
    )
    # Build and return Auth
    return await AuthFactory(client_session).build_auth(
        TEST_EMAIL_ADDRESS, TEST_PASSWORD
    )


@pytest.fixture()
def core_test_object(httpserver: HTTPServer, auth: Auth) -> CoreTestObject:
    """Provide a CoreTestObject for use in component tests."""
    return CoreTestObject(httpserver, auth)



"""Classes related to a DeviceChannel."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Dict

from thermoworks_cloud.utils import parse_datetime, map_firestore_fields, parse_nested_object


@dataclass
class Reading:
    """A temperature reading from a device channel."""

    value: Optional[float] = field(default=None, metadata={
                                   "firestore_type": "doubleValue"})
    """"The temperature units as a string like "F" """
    units: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})


@dataclass
class Alarm:
    """An alarm on a device channel."""

    enabled: Optional[bool] = field(
        default=None, metadata={"firestore_type": "booleanValue"})
    alarming: Optional[bool] = field(
        default=None, metadata={"firestore_type": "booleanValue"})
    value: Optional[int] = field(default=None, metadata={
                                 "firestore_type": "integerValue", "converter": int})
    """"The temperature units as a string like "F" """
    units: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})


@dataclass
class MinMaxReading:
    """A minimum or maximum reading on a device channel."""

    reading: Optional[Reading] = None
    date_reading: Optional[datetime] = field(
        default=None,
        metadata={
            "firestore_type": "timestampValue",
            "converter": parse_datetime
        }
    )


@dataclass
class DeviceChannel:  # pylint: disable=too-many-instance-attributes
    """A device channel on a device.

    All fields are optional as different device types may have different properties.
    """

    last_telemetry_saved: Optional[datetime] = field(
        default=None,
        metadata={
            "firestore_type": "timestampValue",
            "converter": parse_datetime
        }
    )
    """"The last time a telemetry packet was received from the device channel."""
    value: Optional[float] = field(default=None, metadata={
                                   "firestore_type": ["doubleValue","integerValue"]})
    """"The temperature units as a string like "F" """
    units: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    """"The only observed value for this field is "NORMAL"."""
    status: Optional[str] = field(default=None, metadata={
                                  "firestore_type": "stringValue"})
    type: Optional[str] = field(default=None, metadata={
                                "firestore_type": "stringValue"})
    """Customer provided 'name' for this device channel."""
    label: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    last_seen: Optional[datetime] = field(
        default=None,
        metadata={
            "firestore_type": "timestampValue",
            "converter": parse_datetime
        }
    )
    alarm_high: Optional[Alarm] = field(
        default=None, metadata={"api_name": "alarmHigh"})
    alarm_low: Optional[Alarm] = field(
        default=None, metadata={"api_name": "alarmLow"})
    """The device channel number"""
    number: Optional[str] = field(default=None, metadata={
                                  "firestore_type": "stringValue"})
    minimum: Optional[MinMaxReading] = None
    maximum: Optional[MinMaxReading] = None
    show_avg_temp: Optional[bool] = field(
        default=None,
        metadata={
            "api_name": "showAvgTemp",
            "firestore_type": "booleanValue"
        }
    )

    # Dictionary to store any additional properties not explicitly defined
    additional_properties: Optional[Dict] = None


def _parse_min_max_reading(data: dict) -> Optional[MinMaxReading]:
    """Parse minimum or maximum reading data."""
    if not data or "fields" not in data:
        return None

    fields = data["fields"]
    result = MinMaxReading()

    # Parse date_reading
    if "dateReading" in fields and "timestampValue" in fields["dateReading"]:
        result.date_reading = parse_datetime(
            fields["dateReading"]["timestampValue"])

    # Parse reading
    if "reading" in fields and "mapValue" in fields["reading"]:
        result.reading = parse_nested_object(
            fields["reading"]["mapValue"], Reading)

    return result


def _document_to_device_channel(document: dict) -> DeviceChannel:
    """Convert a Firestore Document object into a Device object."""
    fields = document.get("fields", {})
    device_channel = map_firestore_fields(fields, DeviceChannel)

    try:
        # Handle complex objects
        if "alarmHigh" in fields and "mapValue" in fields["alarmHigh"]:
            device_channel.alarm_high = parse_nested_object(
                fields["alarmHigh"]["mapValue"], Alarm)

        if "alarmLow" in fields and "mapValue" in fields["alarmLow"]:
            device_channel.alarm_low = parse_nested_object(
                fields["alarmLow"]["mapValue"], Alarm)

        if "minimum" in fields and "mapValue" in fields["minimum"]:
            device_channel.minimum = _parse_min_max_reading(
                fields["minimum"]["mapValue"])

        if "maximum" in fields and "mapValue" in fields["maximum"]:
            device_channel.maximum = _parse_min_max_reading(
                fields["maximum"]["mapValue"])

    except (KeyError, TypeError, ValueError) as _:
        # If there's an error parsing a specific field, continue with what we have
        pass

    return device_channel



"""Provides capabilities for mocking Thermoworks Cloud related API calls."""

from typing import cast, Dict, Any
from pytest_httpserver import HTTPServer, RequestHandler

from tests.test_data import (
    API_KEY,
    FIREBASE_APPLICATION_BASE_PATH,
    TEST_USER_ID,
)
from thermoworks_cloud.auth import _Auth, Auth


class CoreTestObject:
    """A test object that can be used to mock Thermoworks Cloud related API calls."""

    def __init__(self, httpserver: HTTPServer, auth: Auth) -> None:
        """Create a new test object for the Core module

        Args:
            httpserver (HTTPServer): The http server to use when mocking requests
            auth (Auth): The Auth object to use with this test object
        """
        self.httpserver = httpserver
        self.auth = cast(_Auth, auth)

    def expect_get_user(self, access_token: str) -> RequestHandler:
        """Create a request handler for a GET request to the `get_user` API."""
        url = f"{FIREBASE_APPLICATION_BASE_PATH}/users/{TEST_USER_ID}"
        headers = {"authorization": f"Bearer {access_token}"}
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            url, headers=headers, query_string=query_string
        )

    def expect_get_device(
        self, access_token: str, device_serial: str
    ) -> RequestHandler:
        """Create a request handler for a GET request to the `get_device` API."""
        url = f"{FIREBASE_APPLICATION_BASE_PATH}/devices/{device_serial}"
        headers = {"authorization": f"Bearer {access_token}"}
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            url, headers=headers, query_string=query_string
        )

    def expect_get_device_channel(
        self, access_token: str, device_serial: str, channel: str
    ) -> RequestHandler:
        """Create a request handler for a GET request to the `get_device_channel` API.
        """
        url = f"{FIREBASE_APPLICATION_BASE_PATH}/devices/{device_serial}/channels/{channel}"
        headers = {"authorization": f"Bearer {access_token}"}
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            url, headers=headers, query_string=query_string
        )

    def expect_run_query(
        self, access_token: str, query_body: Dict[str, Any]
    ) -> RequestHandler:
        """Create a request handler for a POST request to the `runQuery` API."""
        url = f"{FIREBASE_APPLICATION_BASE_PATH}:runQuery"
        headers = {
            "authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            url, method="POST", headers=headers, query_string=query_string, json=query_body
        )



"""Test data to be used with mocks"""

import json

# Paths used by the AuthFactory to setup an authentication object
LOGIN_PATH = "/v1/accounts:signInWithPassword"
WEB_CONFIG_PATH = (
    "/v1alpha/projects/-/apps/1:78998049458:web:b41e9d405d8c7de95eefab/webConfig"
)
TOKEN_PATH = "/v1/token"

# Web facing API key for Firebase and Firestore attached to the requests by the AuthFactory
API_KEY = "AIzaSyCf079iccUFc1k7VHdGXng22zXDy8Y3KEY"
TEST_EMAIL_ADDRESS = "test@example.com"
TEST_PROJECT_ID = "test-project-id"
TEST_PASSWORD = "test_password"
TEST_USER_ID = "test_user_id"
TEST_ID_TOKEN = "test_id_token"
TEST_REFRESHED_ID_TOKEN = "test-refreshed-id-token"
TEST_REFRESH_TOKEN = "test_refresh_token"
TEST_ACCOUNT_ID = "test-account-id"

# Root path used for application specific requests
FIREBASE_APPLICATION_BASE_PATH = (
    f"/v1/projects/{TEST_PROJECT_ID}/databases/(default)/documents"
)

# Return values from actual responses
LOGIN_RETURN_VALUE = {
    "idToken": TEST_ID_TOKEN,
    "email": TEST_EMAIL_ADDRESS,
    "refreshToken": TEST_REFRESH_TOKEN,
    "expiresIn": "3600",
    "localId": TEST_USER_ID,
}

# Actual response value structure
CONFIG_RETURN_VALUE = {
    "projectId": TEST_PROJECT_ID,
    "appId": "test-app-id",
    "databaseURL": "https://test-database-url.com",
    "storageBucket": "test-bucket.com",
    "locationId": "test-location-id",
    "authDomain": "test-auth-domain.com",
    "messagingSenderId": "test-message-sender-id",
    "measurementId": "test-measurement-id",
}

# Actual response value structure
TOKEN_REFRESH_RETURN_VALUE = {
    "access_token": TEST_REFRESHED_ID_TOKEN,
    "expires_in": "3600",
    "token_type": "Bearer",
    "refresh_token": "test-refreshed-refresh-token",
    "id_token": TEST_REFRESHED_ID_TOKEN,
    "user_id": TEST_USER_ID,
    "project_id": TEST_PROJECT_ID,
}

TEST_DEVICE_ID_0 = "12:34:56:78:90:AB"
TEST_DEVICE_ID_1 = "CD:EF:12:34:56:78"

# Actual response value structure
USER_RESPONSE = json.loads(
    """
{
  "name": "projects/test-project-name/databases/(default)/documents/users/test-user-id",
  "fields": {
    "system": {
      "mapValue": {
        "fields": {
          "hasPromptedEmailList": {
            "booleanValue": true
          }
        }
      }
    },
    "provider": {
      "stringValue": "firebase"
    },
    "deviceOrder": {
      "mapValue": {
        "fields": {
          "%s": {
            "arrayValue": {
              "values": [
                {
                  "mapValue": {
                    "fields": {
                      "deviceId": {
                        "stringValue": "%s"
                      },
                      "order": {
                        "integerValue": "0"
                      }
                    }
                  }
                },
                {
                  "mapValue": {
                    "fields": {
                      "deviceId": {
                        "stringValue": "%s"
                      },
                      "order": {
                        "integerValue": "1"
                      }
                    }
                  }
                }
              ]
            }
          }
        }
      }
    },
    "notificationSettings": {
      "mapValue": {
        "fields": {
          "emailNotification": {
            "booleanValue": false
          }
        }
      }
    },
    "timeZone": {
      "stringValue": "America/Los_Angeles"
    },
    "appVersion": {
      "stringValue": "1.61.6"
    },
    "displayName": {
      "stringValue": ""
    },
    "email": {
      "stringValue": "%s"
    },
    "exportVersion": {
      "doubleValue": 0.6
    },
    "use24Time": {
      "booleanValue": false
    },
    "emailStatus": {
      "stringValue": "bounce"
    },
    "fcmTokens": {
      "mapValue": {
        "fields": {
          "null": {
            "booleanValue": true
          },
          "token1": {
            "booleanValue": true
          },
          "token2": {
            "booleanValue": true
          }
        }
      }
    },
    "emailLastEvent": {
      "mapValue": {
        "fields": {
          "reason": {
            "stringValue": "Message queue continuously deferred for too long."
          },
          "event": {
            "stringValue": "bounce"
          },
          "email": {
            "stringValue": "%s"
          },
          "bounce_classification": {
            "stringValue": "Unclassified"
          },
          "tls": {
            "integerValue": "0"
          },
          "timestamp": {
            "integerValue": "1234567"
          },
          "smtp-id": {
            "stringValue": "\\u003cmail@server.com\\u003e"
          },
          "type": {
            "stringValue": "blocked"
          },
          "sg_message_id": {
            "stringValue": "12345.0987"
          },
          "sg_event_id": {
            "stringValue": "lkjhg12345"
          }
        }
      }
    },
    "accountId": {
      "stringValue": "test-account-id"
    },
    "uid": {
      "stringValue": "test-uid"
    },
    "roles": {
      "mapValue": {
        "fields": {
          "DataReader": {
            "booleanValue": true
          }
        }
      }
    },
    "lastSeenInApp": {
      "nullValue": null
    },
    "locale": {
      "stringValue": "en-US"
    },
    "photoURL": {
      "stringValue": ""
    },
    "accountRoles": {
      "mapValue": {
        "fields": {
          "accountAdmin": {
            "booleanValue": true
          }
        }
      }
    },
    "preferredUnits": {
      "stringValue": "F"
    },
    "lastLogin": {
      "timestampValue": "2020-01-01T00:00:00.000Z"
    }
  },
  "createTime": "2019-01-01T00:00:00.000Z",
  "updateTime": "2021-01-01T00:00:00.000Z"
}"""
    % (
        TEST_USER_ID,
        TEST_DEVICE_ID_0,
        TEST_DEVICE_ID_1,
        TEST_EMAIL_ADDRESS,
        TEST_EMAIL_ADDRESS,
    )
)

# Actual response value structure
GET_DEVICE_RESPONSE = json.loads(
    """
{
  "name": "projects/test-project-name/databases/(default)/documents/devices/%s",
  "fields": {
    "batteryState": {
      "stringValue": "discharging"
    },
    "bigQuery": {
      "mapValue": {
        "fields": {
          "tableId": {
            "stringValue": "%s"
          },
          "datasetId": {
            "stringValue": "test-dataset-id"
          }
        }
      }
    },
    "battery": {
      "integerValue": "100"
    },
    "lastArchive": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "status": {
      "stringValue": "NORMAL"
    },
    "lastWifiConnection": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "lastBluetoothConnection": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "type": {
      "stringValue": "datalogger"
    },
    "lastTelemetrySaved": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "pendingLoad": {
      "booleanValue": false
    },
    "exportVersion": {
      "doubleValue": 0.5
    },
    "sessionStart": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "deviceId": {
      "stringValue": "%s"
    },
    "serial": {
      "stringValue": "%s"
    },
    "transmitIntervalInSeconds": {
      "integerValue": "7200"
    },
    "deviceDisplayUnits": {
      "stringValue": "F"
    },
    "iotDeviceId": {
      "stringValue": "test-iot-device-id"
    },
    "label": {
      "stringValue": "NODE"
    },
    "lastPurged": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "batteryAlertSent": {
      "booleanValue": false
    },
    "color": {
      "stringValue": "3f90ca"
    },
    "firmware": {
      "stringValue": "1.0.26-26"
    },
    "thumbnail": {
      "stringValue": "node_bl.png"
    },
    "wifi_stength": {
      "integerValue": "-72"
    },
    "recordingIntervalInSeconds": {
      "integerValue": "600"
    },
    "accountId": {
      "stringValue": "test-account-id"
    },
    "lastSeen": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "device": {
      "stringValue": "node"
    }
  },
  "createTime": "2019-01-01T00:00:00.000Z",
  "updateTime": "2021-01-01T00:00:00.000Z"
}"""
    % (TEST_DEVICE_ID_0, TEST_DEVICE_ID_0, TEST_DEVICE_ID_0, TEST_DEVICE_ID_0)
)

# Actual response value structure
GET_DEVICE_CHANNEL_RESPONSE = json.loads(
    """
{
  "name": "projects/test-project-name/databases/(default)/documents/devices/test-device/channels/1",
  "fields": {
    "lastTelemetrySaved": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "value": {
      "doubleValue": -5.1999998092651367
    },
    "color": {
      "stringValue": "none"
    },
    "minimum": {
      "mapValue": {
        "fields": {
          "dateReading": {
            "timestampValue": "2021-01-01T00:00:00.000Z"
          },
          "reading": {
            "mapValue": {
              "fields": {
                "value": {
                  "doubleValue": -9.1000003814697266
                },
                "units": {
                  "stringValue": "F"
                }
              }
            }
          }
        }
      }
    },
    "units": {
      "stringValue": "F"
    },
    "status": {
      "stringValue": "NORMAL"
    },
    "type": {
      "stringValue": "test-device-channel-type"
    },
    "label": {
      "stringValue": "test-channel-label"
    },
    "lastSeen": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "alarmHigh": {
      "mapValue": {
        "fields": {
          "enabled": {
            "booleanValue": true
          },
          "units": {
            "stringValue": "F"
          },
          "alarming": {
            "booleanValue": false
          },
          "value": {
            "integerValue": "30"
          }
        }
      }
    },
    "number": {
      "stringValue": "test-channel-number"
    },
    "maximum": {
      "mapValue": {
        "fields": {
          "reading": {
            "mapValue": {
              "fields": {
                "value": {
                  "doubleValue": 69.5
                },
                "units": {
                  "stringValue": "F"
                }
              }
            }
          },
          "dateReading": {
            "timestampValue": "2021-01-01T00:00:00.000Z"
          }
        }
      }
    },
    "alarmLow": {
      "mapValue": {
        "fields": {
          "enabled": {
            "booleanValue": true
          },
          "units": {
            "stringValue": "F"
          },
          "alarming": {
            "booleanValue": false
          },
          "value": {
            "integerValue": "-12"
          }
        }
      }
    },
    "showAvgTemp": {
      "booleanValue": true
    }
  },
  "createTime": "2019-01-01T00:00:00.000Z",
  "updateTime": "2021-01-01T00:00:00.000Z"
}"""
)

GET_DEVICE_CHANNEL_RESPONSE_INT = json.loads(
    """
{
  "name": "projects/test-project-name/databases/(default)/documents/devices/test-device/channels/1",
  "fields": {
    "lastTelemetrySaved": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "value": {
      "integerValue": -6
    },
    "color": {
      "stringValue": "none"
    },
    "minimum": {
      "mapValue": {
        "fields": {
          "dateReading": {
            "timestampValue": "2021-01-01T00:00:00.000Z"
          },
          "reading": {
            "mapValue": {
              "fields": {
                "value": {
                  "doubleValue": -9.1000003814697266
                },
                "units": {
                  "stringValue": "F"
                }
              }
            }
          }
        }
      }
    },
    "units": {
      "stringValue": "F"
    },
    "status": {
      "stringValue": "NORMAL"
    },
    "type": {
      "stringValue": "test-device-channel-type"
    },
    "label": {
      "stringValue": "test-channel-label"
    },
    "lastSeen": {
      "timestampValue": "2021-01-01T00:00:00.000Z"
    },
    "alarmHigh": {
      "mapValue": {
        "fields": {
          "enabled": {
            "booleanValue": true
          },
          "units": {
            "stringValue": "F"
          },
          "alarming": {
            "booleanValue": false
          },
          "value": {
            "integerValue": "30"
          }
        }
      }
    },
    "number": {
      "stringValue": "test-channel-number"
    },
    "maximum": {
      "mapValue": {
        "fields": {
          "reading": {
            "mapValue": {
              "fields": {
                "value": {
                  "doubleValue": 69.5
                },
                "units": {
                  "stringValue": "F"
                }
              }
            }
          },
          "dateReading": {
            "timestampValue": "2021-01-01T00:00:00.000Z"
          }
        }
      }
    },
    "alarmLow": {
      "mapValue": {
        "fields": {
          "enabled": {
            "booleanValue": true
          },
          "units": {
            "stringValue": "F"
          },
          "alarming": {
            "booleanValue": false
          },
          "value": {
            "integerValue": "-12"
          }
        }
      }
    },
    "showAvgTemp": {
      "booleanValue": true
    }
  },
  "createTime": "2019-01-01T00:00:00.000Z",
  "updateTime": "2021-01-01T00:00:00.000Z"
}"""
)

# Response for get_devices query - using sanitized real-world format
GET_DEVICES_RESPONSE = [
    {
        "document": {
            "name": f"projects/test-project-name/databases/(default)/documents/devices/{TEST_DEVICE_ID_0}",  # pylint: disable=line-too-long
            "fields": {
                "label": {"stringValue": "NODE"},
                "device": {"stringValue": "node"},
                "deviceId": {"stringValue": TEST_DEVICE_ID_0},
                "lastPurged": {"timestampValue": "2023-04-06T00:36:54.483Z"},
                "exportVersion": {"doubleValue": 0.5},
                "serial": {"stringValue": TEST_DEVICE_ID_0},
                "batteryState": {"stringValue": "discharging"},
                "accountId": {"stringValue": TEST_ACCOUNT_ID},
                "sessionStart": {"timestampValue": "2023-09-23T16:36:09.435Z"},
                "pendingLoad": {"booleanValue": False},
                "firmware": {"stringValue": "1.0.26-26"},
                "bigQuery": {
                    "mapValue": {
                        "fields": {
                            "tableId": {"stringValue": TEST_DEVICE_ID_0},
                            "datasetId": {"stringValue": "test-dataset-id"}
                        }
                    }
                },
                "deviceDisplayUnits": {"stringValue": "F"},
                "thumbnail": {"stringValue": "node_bl.png"},
                "color": {"stringValue": "3f90ca"},
                "recordingIntervalInSeconds": {"integerValue": "600"},
                "iotDeviceId": {"stringValue": "T123456"},
                "transmitIntervalInSeconds": {"integerValue": "7200"},
                "battery": {"integerValue": "100"},
                "batteryAlertSent": {"booleanValue": False},
                "status": {"stringValue": "NORMAL"},
                "lastBluetoothConnection": {"timestampValue": "2021-01-01T00:00:00.000Z"},
                "type": {"stringValue": "datalogger"},
                "lastArchive": {"timestampValue": "2021-01-01T00:00:00.000Z"},
                "wifi_stength": {"integerValue": "-72"},
                "lastTelemetrySaved": {"timestampValue": "2021-01-01T00:00:00.000Z"},
                "lastSeen": {"timestampValue": "2021-01-01T00:00:00.000Z"},
                "lastWifiConnection": {"timestampValue": "2021-01-01T00:00:00.000Z"}
            },
            "createTime": "2019-01-01T00:00:00.000Z",
            "updateTime": "2021-01-01T00:00:00.000Z"
        },
        "readTime": "2023-01-01T00:00:00.000000000Z"
    },
    {
        "document": {
            "name": f"projects/test-project-name/databases/(default)/documents/devices/{TEST_DEVICE_ID_1}",  # pylint: disable=line-too-long
            "fields": {
                "deviceId": {"stringValue": TEST_DEVICE_ID_1},
                "serial": {"stringValue": TEST_DEVICE_ID_1},
                "label": {"stringValue": "SIGNALS"},
                "type": {"stringValue": "signals"},
                "accountId": {"stringValue": TEST_ACCOUNT_ID},
                "device": {"stringValue": "signals"},
                "status": {"stringValue": "NORMAL"},
                "lastSeen": {"timestampValue": "2021-01-01T00:00:00.000Z"}
            },
            "createTime": "2019-01-01T00:00:00.000Z",
            "updateTime": "2021-01-01T00:00:00.000Z"
        },
        "readTime": "2023-01-01T00:00:00.000000000Z"
    }
]



"""Utility functions used within the library."""

from dataclasses import Field, fields as dataclass_fields
from datetime import datetime
from typing import Any, Callable, Dict, Optional, Type

from aiohttp import ClientResponse


def parse_datetime(value: str) -> datetime:
    """Convert Firestore timestamp string to a datetime object."""
    return datetime.fromisoformat(value)


def unwrap_firestore_value(value_dict):
    """Unwrap a Firestore value dictionary into a single Python value.

    Args:
        value_dict (dict): A Firestore value dictionary containing a type and value

    Returns:
        The Python value
    """
    value = value_dict.values()
    if len(value) != 1:
        raise ValueError("Firestore values must contain a single value")
    return next(iter(value))


def get_field_value(fields: Dict, field_name: str, value_type: str,
                    converter: Optional[Callable] = None) -> Any:
    """Helper function to safely get values from Firestore fields.

    Args:
        fields: Dictionary containing Firestore fields
        field_name: Name of the field to retrieve
        value_type: Type of value to retrieve (e.g., "stringValue", "integerValue")
        converter: Optional function to convert the value

    Returns:
        The value if found, or None if not found
    """
    if field_name in fields and value_type in fields[field_name]:
        value = fields[field_name][value_type]
        return converter(value) if converter else value
    return None


def api_field_name(field: Field) -> str:
    """Get the API field name for a dataclass field.

    Checks for a 'api_name' metadata entry, otherwise converts from snake_case to camelCase.

    Args:
        field: A dataclass field

    Returns:
        The API field name
    """
    # Check if the field has an api_name metadata entry
    if 'api_name' in field.metadata:
        return field.metadata['api_name']

    # Otherwise convert from snake_case to camelCase
    parts = field.name.split('_')
    return parts[0] + ''.join(p.capitalize() for p in parts[1:])


def extract_additional_properties(firestore_fields: Dict, dataclass_type: Type) -> Optional[Dict]:
    """Extract additional properties not explicitly mapped.

    Args:
        firestore_fields: Dictionary containing Firestore fields
        dataclass_type: The dataclass type to extract field names from

    Returns:
        Dictionary of additional properties, or None if none found
    """
    # Generate known fields from dataclass
    known_fields = set()
    for field in dataclass_fields(dataclass_type):
        if field.name == 'additional_properties':
            continue

        # Get the API field name
        known_fields.add(api_field_name(field))

    # Extract additional properties
    additional_props = {}
    for field_name, field_value in firestore_fields.items():
        if field_name not in known_fields:
            # Store the raw value for additional properties
            value_type = next(iter(field_value.keys()))
            additional_props[field_name] = field_value[value_type]

    return additional_props if additional_props else None


def parse_map_field(field_value: Dict, value_type: str) -> Optional[Dict]:
    """Parse a map field into a dictionary.

    Args:
        field_value: The Firestore field value containing a mapValue
        value_type: The type of values in the map (e.g., "booleanValue")

    Returns:
        A dictionary with the parsed values
    """
    if "mapValue" not in field_value or "fields" not in field_value["mapValue"]:
        return None

    result = {}
    for key, value in field_value["mapValue"]["fields"].items():
        if value_type in value:
            result[key] = value[value_type]

    return result


def map_firestore_fields(firestore_fields: Dict, dataclass_type: Type) -> Any:
    """Map Firestore fields to a dataclass instance based on field metadata.

    Args:
        firestore_fields: Dictionary containing Firestore fields
        dataclass_type: The dataclass type to map fields to

    Returns:
        An instance of the dataclass with fields populated from Firestore data
    """
    instance = dataclass_type()

    for field_info in dataclass_fields(dataclass_type):
        if field_info.name == 'additional_properties':
            continue

        # Get API field name
        api_name = api_field_name(field_info)

        # Skip if field is not in Firestore data
        if api_name not in firestore_fields:
            continue

        # Handle simple fields with firestore_type metadata
        if 'firestore_type' in field_info.metadata:
            firestore_type = field_info.metadata['firestore_type']
            converter = field_info.metadata.get('converter')

            if isinstance(firestore_type, list):
                for type_item in firestore_type:
                    value = get_field_value(
                        firestore_fields, api_name, type_item, converter)
                    if value is not None:
                        setattr(instance, field_info.name, value)
                        break
            else:
                # Set the field value
                value = get_field_value(
                    firestore_fields, api_name, firestore_type, converter)
                setattr(instance, field_info.name, value)

        # Handle map fields with map_of metadata
        elif 'map_of' in field_info.metadata:
            value_type = field_info.metadata['map_of']
            value = parse_map_field(firestore_fields[api_name], value_type)
            setattr(instance, field_info.name, value)

    # Handle additional properties
    instance.additional_properties = extract_additional_properties(
        firestore_fields, dataclass_type)

    return instance


def parse_nested_object(data: dict, dataclass_type: Type) -> Any:
    """Parse a nested Firestore object into a dataclass instance.

    Args:
        data: The Firestore data containing a mapValue with fields
        dataclass_type: The dataclass type to map the fields to

    Returns:
        An instance of the dataclass type, or None if the data is invalid
    """
    if not data or "fields" not in data:
        return None

    return map_firestore_fields(data["fields"], dataclass_type)


async def format_client_response(response: ClientResponse) -> str:
    """Format a string from the pertinent details of a response."""
    return f"status={response.status} reason={response.reason} body={await response.text()}"



"""Test suite for the Auth object."""

from aiohttp import ClientSession, ClientResponseError
import pytest

from tests.auth_test_object import AuthTestObject
from tests.test_data import (
    CONFIG_RETURN_VALUE,
    LOGIN_RETURN_VALUE,
    TEST_EMAIL_ADDRESS,
    TEST_PASSWORD,
    TEST_USER_ID,
    TOKEN_REFRESH_RETURN_VALUE,
)
from thermoworks_cloud import (
    AuthFactory,
    AuthenticationError,
    AuthenticationErrorReason,
)


class TestAuth:
    """Test the Auth object."""

    async def test_build_auth(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test the happy path for building an auth object."""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
        auth_test_object.expect_login(
            TEST_EMAIL_ADDRESS, TEST_PASSWORD
        ).respond_with_json(LOGIN_RETURN_VALUE)

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        auth = await auth_factory.build_auth(TEST_EMAIL_ADDRESS, TEST_PASSWORD)
        assert auth.user_id == TEST_USER_ID

    async def test_get_config_4xx(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that 4xx errors are swallowed with the error cause captured."""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_data(
            status=400, response_data="invalid request"
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth(TEST_EMAIL_ADDRESS, TEST_PASSWORD)
        except RuntimeError as e:
            # Validate that the underlying client response error is available for debugging
            assert isinstance(e.__cause__, ClientResponseError)

    async def test_get_config_5xx(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that 5xx errors are swallowed with the error cause captured."""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_data(
            status=500, response_data="internal server error"
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth(TEST_EMAIL_ADDRESS, TEST_PASSWORD)
        except RuntimeError as e:
            # Validate that the underlying client response error is available for debugging
            assert isinstance(e.__cause__, ClientResponseError)

    async def test_login_5xx(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that 5xx errors are swallowed with the error cause captured."""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
        auth_test_object.expect_login(
            TEST_EMAIL_ADDRESS, TEST_PASSWORD
        ).respond_with_data(status=500, response_data="internal error")

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth(TEST_EMAIL_ADDRESS, TEST_PASSWORD)
        except RuntimeError as e:
            # Validate that the underlying client response error is available for debugging
            assert isinstance(e.__cause__, ClientResponseError)

    async def test_invalid_email(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that an invaid email response is converted to an AuthenticationError."""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
        auth_test_object.expect_login("invalid-email", TEST_PASSWORD).respond_with_json(
            status=400,
            response_json={
                # From actual request
                "error": {
                    "code": 400,
                    "message": "INVALID_EMAIL",
                    "errors": [
                        {
                            "message": "INVALID_EMAIL",
                            "domain": "global",
                            "reason": "invalid",
                        }
                    ],
                }
            },
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth("invalid-email", TEST_PASSWORD)
        except AuthenticationError as e:
            assert e.reason == AuthenticationErrorReason.INVALID_EMAIL

    async def test_email_not_found(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that an unknown email address is converted to an AuthenticationError."""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
        auth_test_object.expect_login(
            "unknown@example.com", TEST_PASSWORD
        ).respond_with_json(
            status=400,
            response_json={
                # From actual response
                "error": {
                    "code": 400,
                    "message": "EMAIL_NOT_FOUND",
                    "errors": [
                        {
                            "message": "EMAIL_NOT_FOUND",
                            "domain": "global",
                            "reason": "invalid",
                        }
                    ],
                }
            },
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth("unknown@example.com", TEST_PASSWORD)
        except AuthenticationError as e:
            assert e.reason == AuthenticationErrorReason.EMAIL_NOT_FOUND

    async def test_invalid_password(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that an invalid password response is converted to an AuthenticationError"""

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
        auth_test_object.expect_login(
            TEST_EMAIL_ADDRESS, "invalid-password"
        ).respond_with_json(
            status=400,
            response_json={
                "error": {
                    "code": 400,
                    "message": "INVALID_PASSWORD",
                    "errors": [
                        {
                            "message": "INVALID_PASSWORD",
                            "domain": "global",
                            "reason": "invalid",
                        }
                    ],
                }
            },
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth(TEST_EMAIL_ADDRESS, "invalid-password")
        except AuthenticationError as e:
            assert e.reason == AuthenticationErrorReason.INVALID_PASSWORD

    async def test_unknown_error(
            self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that a response with an invalid error reason is converted 
        to an AuthenticationError.
        """

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)
        auth_test_object.expect_login(
            TEST_EMAIL_ADDRESS, TEST_PASSWORD
        ).respond_with_json(
            status=400,
            response_json={
                "error": {
                    "code": 400,
                    "message": "UKNOWN_ERROR_MESSAGE",
                    "errors": [
                        {
                            "message": "UKNOWN_ERROR_MESSAGE",
                            "domain": "global",
                            "reason": "unknown-reason",
                        }
                    ],
                }
            },
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        try:
            await auth_factory.build_auth(TEST_EMAIL_ADDRESS, TEST_PASSWORD)
        except AuthenticationError as e:
            assert e.reason == AuthenticationErrorReason.UNKNOWN

    def test_authfactory_none_websession(self):
        """Test that AuthFactory cannot be created without a ClientSession."""

        with pytest.raises(ValueError):
            AuthFactory(websession=None)

    async def test_request_refresh_token(
        self, auth_test_object: AuthTestObject, client_session: ClientSession
    ):
        """Test that the Auth object correctly refreshes the access token when 
        executing a request.
        """

        # Setup mock responses
        auth_test_object.expect_config().respond_with_json(CONFIG_RETURN_VALUE)

        # if the token expires in less than 60 seconds it should get refreshed
        # when `async_get_access_token` is called
        auth_test_object.expect_login(
            TEST_EMAIL_ADDRESS, TEST_PASSWORD
        ).respond_with_json(LOGIN_RETURN_VALUE | {"expiresIn": "10"})
        auth_test_object.expect_token_refresh(
            LOGIN_RETURN_VALUE["refreshToken"]
        ).respond_with_json(TOKEN_REFRESH_RETURN_VALUE)
        headers = {
            "Authorization": f"Bearer {TOKEN_REFRESH_RETURN_VALUE['access_token']}",
        }
        auth_test_object.expect_request("test", headers=headers).respond_with_json(
            {"test": "test"}
        )

        # act
        auth_factory = AuthFactory(client_session)

        # assert
        auth = await auth_factory.build_auth(TEST_EMAIL_ADDRESS, TEST_PASSWORD)
        respose = await auth.request("GET", "documents/test")
        assert await respose.json() == {"test": "test"}



"""Demonstrates how to use this library to get information about the logged in user.

This example shows how to:
1. Authenticate with the Thermoworks Cloud API
2. Get user information
3. Print detailed information about the user and their properties
"""

import asyncio
import json
import os
import pprint
from dataclasses import asdict

from aiohttp import ClientSession

from thermoworks_cloud import AuthFactory, ThermoworksCloud

# Make sure these are defined
email = os.environ["THERMOWORKS_EMAIL"]
password = os.environ["THERMOWORKS_PASSWORD"]


async def __main__():
    # Use a context manager when providing the session to the auth factory
    async with ClientSession() as session:
        auth = await AuthFactory(session).build_auth(email, password)
        thermoworks = ThermoworksCloud(auth)
        user = await thermoworks.get_user()

        # Print user information
        print(f"\n{'=' * 50}")
        print(f"USER: {user.display_name or user.email or 'Unknown'}")
        print(f"UID: {user.uid}")
        print(f"ACCOUNT ID: {user.account_id}")
        print(f"{'=' * 50}")

        # Print user properties
        print("\nUSER PROPERTIES:")
        user_dict = asdict(user)
        pprint.pprint(user_dict, compact=False)

        # Print device count
        device_count = 0
        if user.device_order and user.account_id in user.device_order:
            device_count = len(user.device_order[user.account_id])
        print(f"\nDEVICES: {device_count}")

        # Save the data to a JSON file for further analysis
        with open("thermoworks_user_data.json", "w") as f:
            json.dump(user_dict, f, indent=2, default=str)

        print("\nData saved to thermoworks_user_data.json for further analysis")


asyncio.run(__main__())



"""Accessor for the Thermoworks Cloud API."""

import logging
from typing import List

from thermoworks_cloud.utils import format_client_response

from .auth import Auth
from .models.device import Device, _document_to_device
from .models.device_channel import DeviceChannel, _document_to_device_channel
from .models.user import User, document_to_user

_LOGGER = logging.getLogger(__name__)


class ResourceNotFoundError(Exception):
    """Custom exception indicating that the requested resource could not be found."""

    def __init__(
        self,
        message: str,
    ) -> None:
        """Initialize the error."""

        super().__init__(message)
        self.message = message


class ThermoworksCloud:
    """Client for the Thermoworks Cloud service."""

    def __init__(self, auth: Auth) -> None:
        """Create a new client. `thermoworks_cloud.Auth` objects are created using a
        `thermoworks_cloud.AuthFactory`.

        Args:
            auth (Auth): Authorization object used to make authorized requests to the service.
        """
        self._auth = auth

    async def get_user(self) -> User:
        """Fetch information for the authenticated user."""

        response = await self._auth.request("get", f"documents/users/{self._auth.user_id}")
        if response.ok:
            user_document = await response.json()
            return document_to_user(user_document)

        if response.status == 404:
            raise ResourceNotFoundError("User not found")

        try:
            error_response = await format_client_response(response)
        except RuntimeError:
            error_response = "Could not read response body."

        raise RuntimeError(f"Failed to get user: {error_response}")

    async def get_device(self, device_serial: str) -> Device:
        """Fetch a device by serial number."""

        response = await self._auth.request("get", f"documents/devices/{device_serial}")
        if response.ok:
            device_document = await response.json()
            return _document_to_device(device_document)

        if response.status == 404:
            raise ResourceNotFoundError(f"Device with serial '{device_serial}' not found")

        try:
            error_response = await format_client_response(response)
        except RuntimeError:
            error_response = "Could not read response body."

        raise RuntimeError(f"Failed to get device: {error_response}")

    async def get_device_channel(
        self, device_serial: str, channel: str
    ) -> DeviceChannel:
        """Fetch channel information for a device."""

        response = await self._auth.request(
            "get", f"documents/devices/{device_serial}/channels/{channel}"
        )
        if response.ok:
            device_channel_document = await response.json()
            return _document_to_device_channel(device_channel_document)

        if response.status == 404:
            raise ResourceNotFoundError(
                f"Device channel with serial '{device_serial}' and channel '{channel}' not found"
            )

        try:
            error_response = await format_client_response(response)
        except RuntimeError:
            error_response = "Could not read response body."

        raise RuntimeError(
            f"Failed to get device channel: {error_response}")

    async def get_devices(self, account_id: str) -> List[Device]:
        """Fetch devices for the authenticated user using Firestore query.

        It queries the Firestore 'devices' collection directly using the user's accountId.
        """

        # Build the Firestore query
        query_body = {
            "structuredQuery": {
                "from": [{"collectionId": "devices"}],
                "where": {
                    "fieldFilter": {
                        "field": {"fieldPath": "accountId"},
                        "op": "EQUAL",
                        "value": {"stringValue": account_id}
                    }
                },
                "orderBy": [{"field": {"fieldPath": "__name__"}, "direction": "ASCENDING"}]
            }
        }

        # Execute the query against Firestore REST API
        headers = {"Content-Type": "application/json"}
        response = await self._auth.request("post", "documents:runQuery",
                                            additional_headers=headers, json=query_body)

        if not response.ok:
            try:
                error_response = await format_client_response(response)
            except RuntimeError:
                error_response = "Could not read response body."
            raise RuntimeError(f"Failed to query devices: {error_response}")

        # Process the response
        result = await response.json()
        devices = []

        for item in result:
            if "document" in item:
                device = _document_to_device(item["document"])
                devices.append(device)

        return devices



"""Manage authentication for the Thermoworks Cloud API."""

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from typing import Protocol, TypedDict, cast

from aiohttp import ClientResponse, ClientResponseError, ClientSession

from .models.user_credentials import _UserCredentials, _UserLoginResponse


class Auth(Protocol):
    """Interface to make authenticated HTTP requests to the ThermoWorks Cloud service."""

    @property
    def user_id(self) -> str:
        """The id of the logged in user"""
        ...  # pylint: disable=unnecessary-ellipsis

    async def request(self, method, url,  additional_headers=None, json=None) -> ClientResponse:
        """Make an authenticated request."""
        ...  # pylint: disable=unnecessary-ellipsis


class _AuthBase(ABC, Auth):
    """Abstract class to make authenticated requests."""

    def __init__(self, websession: ClientSession, host: str, api_key: str) -> None:
        """Initialize the auth."""
        self.websession = websession
        self.host = host
        self.api_key = api_key

    @abstractmethod
    async def _async_get_access_token(self) -> str:
        """Return a valid access token."""
        ...  # pylint: disable=unnecessary-ellipsis

    async def request(self, method, url, additional_headers=None, json=None) -> ClientResponse:

        access_token = await self._async_get_access_token()
        headers = {"authorization": f"Bearer {access_token}"}

        # If headers are provided in kwargs, merge them with our auth headers
        if additional_headers:
            headers.update(additional_headers)

        url = f"{self.host}/{url}?key={self.api_key}"

        return await self.websession.request(
            method,
            url,
            headers=headers,
            json=json
        )


def _is_expired(expiration_time, buffer_seconds=60):
    """Determine if the key needs renewal based on a buffer time.

    Args:
        expiration_time (datetime): The calculated expiration time of the key.
        buffer_seconds (int): Time before expiration to renew the key (default is 60 seconds).

    Returns:
        bool: True if the key needs renewal, False otherwise.

    """
    # Get the current time
    current_time = datetime.now()
    experation_threshold = expiration_time - timedelta(seconds=buffer_seconds)

    # Check if the key is within the renewal buffer time
    return current_time >= experation_threshold


class _WebConfigResponse(TypedDict):
    """See
    <https://firebase.google.com/docs/reference/firebase-management/rest/v1beta1/projects>.
    """

    projectId: str
    appId: str
    databaseURL: str
    storageBucket: str
    locationId: str
    authDomain: str
    messagingSenderId: str
    measurementId: str


class FirestoreError(TypedDict):
    """See <https://firebase.google.com/docs/reference/rest/auth#section-error-format>."""

    message: str
    domain: str
    reason: str


class ErrorResponse(TypedDict):
    """See <https://firebase.google.com/docs/reference/rest/auth#section-error-format>."""

    code: int
    message: str
    errors: list[FirestoreError]


class AuthenticationErrorReason(Enum):
    """See
    <https://firebase.google.com/docs/reference/rest/auth?authuser=0#section-sign-in-email-password>.
    """

    INVALID_EMAIL = "INVALID_EMAIL"
    EMAIL_NOT_FOUND = "EMAIL_NOT_FOUND"
    INVALID_PASSWORD = "INVALID_PASSWORD"
    USER_DISABLED = "USER_DISABLED"
    UNKNOWN = "UNKNOWN"
    """Catch-all for any undocumented error reasons"""


class AuthenticationError(Exception):
    """Custom exception for authentication errors."""

    def __init__(
        self,
        message: str,
        reason: AuthenticationErrorReason,
        details: list[FirestoreError],
    ) -> None:
        """Initialize an authentication error."""

        super().__init__(
            f"Error authenticating with Firestore: {reason.value}")
        self.message = message
        self.reason = reason
        self.details = details


class _TokenManager:
    """Manage access tokens for Auth."""

    _IDENTITY_HOST = "https://identitytoolkit.googleapis.com"
    _TOKEN_HOST = "https://securetoken.googleapis.com"

    _user_credentials: _UserCredentials

    def __init__(self, websession: ClientSession, api_key: str) -> None:
        """Initialize token manager."""
        self._websession = websession
        self._api_key = api_key

    async def login(self, email: str, password: str) -> None:
        """Exchange login credentials for token credentials."""

        url = f"{self._IDENTITY_HOST}/v1/accounts:signInWithPassword"
        query = {"key": self._api_key}
        headers = {
            "Content-Type": "application/json",
            "referer": "https://cloud.thermoworks.com/"
        }
        json = {
            "email": email,
            "password": password,
            "returnSecureToken": True,
        }

        response = await self._websession.request(
            "post", url, headers=headers, json=json, params=query
        )

        if response.ok:
            login_response = await response.json()
            cast(_UserLoginResponse, login_response)
            self._user_credentials = _UserCredentials.from_user_login_response(
                login_response
            )
        elif response.status == 400:
            login_response = await response.json()
            if login_response["error"]:
                error_response = cast(ErrorResponse, login_response["error"])
                error_message = error_response["message"]
                error_details = error_response["errors"]

                try:
                    error_reason = AuthenticationErrorReason(error_message)
                except ValueError:
                    error_reason = AuthenticationErrorReason.UNKNOWN
                raise AuthenticationError(
                    message=f"Authentication failed: {error_message}",
                    reason=error_reason,
                    details=error_details,
                )
        else:
            try:
                response.raise_for_status()
            except ClientResponseError as e:
                raise RuntimeError("Unable to authenticate") from e

    @property
    def user_id(self) -> str:
        """Return id of the user that the manager is authenticate with."""
        return self._user_credentials.user_id

    @property
    def access_token(self) -> str:
        """Return an access token for use with Auth."""
        return self._user_credentials.access_token

    def is_token_valid(self) -> bool:
        """Return a bool indicating whether the token is expired."""

        return not _is_expired(self._user_credentials.expiration_time)

    async def refresh_access_token(self) -> None:
        """Refresh the access token."""

        url = f"{self._TOKEN_HOST}/v1/token?key={self._api_key}"
        headers = {
            "Content-Type": "application/json",
            "referer": "https://cloud.thermoworks.com/"
        }
        json = {
            "grant_type": "refresh_token",
            "refresh_token": self._user_credentials.refresh_token,
        }

        response = await self._websession.request(
            "post", url, headers=headers, json=json
        )
        response.raise_for_status()
        refresh_token_response = await response.json()
        self._user_credentials = _UserCredentials.from_refresh_token_response(
            refresh_token_response
        )


class _Auth(_AuthBase):
    """Execute authenticated requests."""

    def __init__(
        self,
        websession: ClientSession,
        api_url_root: str,
        api_key: str,
        token_manager: _TokenManager,
    ) -> None:
        """Initialize the auth."""
        super().__init__(websession, api_url_root, api_key)
        self.token_manager = token_manager

    @property
    def user_id(self) -> str:
        """The id of the user that is authenticated."""
        return self.token_manager.user_id

    async def _async_get_access_token(self) -> str:
        """Return a valid access token."""
        if self.token_manager.is_token_valid():
            return self.token_manager.access_token

        await self.token_manager.refresh_access_token()
        return self.token_manager.access_token


class AuthFactory:  # pylint: disable=too-few-public-methods
    """Builds `thermoworks_cloud.Auth` objects."""

    _API_KEY = "AIzaSyCf079iccUFc1k7VHdGXng22zXDy8Y3KEY"
    _APP_ID = "1:78998049458:web:b41e9d405d8c7de95eefab"
    _FIREBASE_HOST = "https://firebase.googleapis.com"
    _FIRESTORE_HOST = "https://firestore.googleapis.com"

    def __init__(self, websession: ClientSession) -> None:
        """Initialize the auth factory.

        Args:
            websession (ClientSession): The HTTP client to be used when authenticating the user.
        """

        if websession is None:
            raise ValueError("parameter cannot be None")

        self._websession = websession

    async def _get_config(self) -> _WebConfigResponse:
        """Get the Firestore project information for this application."""

        url = f"{self._FIREBASE_HOST}/v1alpha/projects/-/apps/{self._APP_ID}/webConfig"
        headers = {
            "accept": "application/json",
            "x-goog-api-key": self._API_KEY,
            "referer": "https://cloud.thermoworks.com/"
        }

        try:
            response = await self._websession.request("get", url, headers=headers)
            response.raise_for_status()
            return await response.json()
        except Exception as e:
            raise RuntimeError(
                "Unable to fetch application configuration") from e

    async def build_auth(self, email: str, password: str) -> Auth:
        """Build an Auth instance.

        Args:
            email (str): The email address of a ThermoWorks Cloud user.
            password (str): The password for the user.

        Returns:
            Auth: An Auth object for making requests on behalf of the authenticated user.
        """

        web_config = await self._get_config()
        project_id = web_config["projectId"]
        url_root = f"{self._FIRESTORE_HOST}/v1/projects/{project_id}/databases/(default)"

        token_manager = _TokenManager(self._websession, self._API_KEY)
        await token_manager.login(email, password)
        return _Auth(
            self._websession,
            api_url_root=url_root,
            api_key=self._API_KEY,
            token_manager=token_manager,
        )



"""Test suite for the Core Thermoworks object."""

from datetime import datetime
from typing import Any
from unittest.mock import AsyncMock, patch
import pytest

from tests.test_data import GET_DEVICE_CHANNEL_RESPONSE_INT
from tests.core_test_object import CoreTestObject
from tests.test_data import (
    TEST_DEVICE_ID_0,
    TEST_DEVICE_ID_1,
    TEST_EMAIL_ADDRESS,
    TEST_ID_TOKEN,
    TEST_USER_ID,
    TEST_ACCOUNT_ID,
    USER_RESPONSE,
    GET_DEVICE_RESPONSE,
    GET_DEVICE_CHANNEL_RESPONSE,
    GET_DEVICES_RESPONSE,
)
from thermoworks_cloud.auth import Auth
from thermoworks_cloud import ThermoworksCloud


def iso_instant(datetime_input: datetime) -> str:
    """Convert a datetime object to an ISO 8601 formatted 'instant' string.

    Args:
        datetime_input (datetime): The datetime object to convert.

    Returns:
        str: The ISO 8601 formatted string representation of the datetime object.
    """
    return datetime_input.isoformat(timespec="milliseconds").replace("+00:00", "Z")


def get_value(value_object: dict) -> Any:
    """Get the value from a nested dictionary.

    Args:
        value_object: The dictionary to search in

    Returns:
        The value at the specified field path
    """
    assert len(value_object) == 1, "Value object can only contain a single value"
    return list(value_object.values())[0]


def get_field_value(response_dict: dict, field_path: str):
    """Get a value from a nested dictionary using a field path.

    Args:
        response_dict: The dictionary to search in
        field_path: The path to the field, separated by dots

    Returns:
        The value at the specified field path
    """
    field_path_segments = field_path.split(".")
    fields = response_dict["fields"]
    for field_path_segment in field_path_segments:
        fields = fields[field_path_segment]

    return get_value(fields)


def assert_map_values(response_dict: dict, obj_dict: dict, field_path: str):
    """Compare boolean values from nested response dictionary with object dictionary

    Args:
        response_dict: The response dictionary containing mapValue fields
        obj_dict: The object dictionary to compare against
        field_path: The field path in response_dict to check
    """
    fields = response_dict["fields"][field_path]["mapValue"]["fields"]
    for key, value in fields.items():
        assert obj_dict[key] == get_value(value)


class TestCore:  # pylint: disable=too-many-public-methods
    """Test class for ThermoworksCloud core functionality."""

    async def test_get_user(self, auth: Auth, core_test_object: CoreTestObject):
        """Test the get_user method of ThermoworksCloud.

        This test checks if the get_user method returns the expected User object
        when a valid user_id is provided.
        """

        # Setup
        core_test_object.expect_get_user(access_token=TEST_ID_TOKEN).respond_with_json(
            USER_RESPONSE
        )
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        user = await thermoworks_cloud.get_user()

        # Assert
        assert user is not None
        assert user.uid == get_field_value(USER_RESPONSE, "uid")
        assert user.display_name == get_field_value(
            USER_RESPONSE, "displayName")
        assert user.email == get_field_value(USER_RESPONSE, "email")
        assert user.provider == get_field_value(USER_RESPONSE, "provider")
        assert user.time_zone == get_field_value(USER_RESPONSE, "timeZone")
        assert user.app_version == get_field_value(USER_RESPONSE, "appVersion")
        assert user.preferred_units == get_field_value(
            USER_RESPONSE, "preferredUnits")
        assert user.locale == get_field_value(USER_RESPONSE, "locale")
        assert user.photo_url == get_field_value(USER_RESPONSE, "photoURL")
        assert user.use_24_time == get_field_value(USER_RESPONSE, "use24Time")
        assert_map_values(USER_RESPONSE, user.roles, "roles")
        assert_map_values(USER_RESPONSE, user.account_roles, "accountRoles")
        assert_map_values(USER_RESPONSE, user.system, "system")
        assert_map_values(
            USER_RESPONSE, user.notification_settings, "notificationSettings"
        )
        assert_map_values(USER_RESPONSE, user.fcm_tokens, "fcmTokens")

        # Devices
        device_order_items = user.device_order[TEST_USER_ID]
        assert len(device_order_items) == 2
        assert device_order_items[0].device_id == TEST_DEVICE_ID_0
        assert device_order_items[1].device_id == TEST_DEVICE_ID_1
        assert user.email_last_event is not None
        assert user.email_last_event.email == TEST_EMAIL_ADDRESS
        assert user.export_version == get_field_value(
            USER_RESPONSE, "exportVersion")
        assert user.last_seen_in_app == get_field_value(
            USER_RESPONSE, "lastSeenInApp")
        assert iso_instant(user.last_login) == get_field_value(
            USER_RESPONSE, "lastLogin"
        )
        assert iso_instant(user.create_time) == USER_RESPONSE["createTime"]
        assert iso_instant(user.update_time) == USER_RESPONSE["updateTime"]

    async def test_get_user_4xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_user method of ThermoworksCloud when a 4xx error is returned.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_user method raises an exception when a 4xx error is returned.
        """
        # Setup
        core_test_object.expect_get_user(access_token=TEST_ID_TOKEN).respond_with_data(
            status=400, response_data=b"Bad Request"
        )
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_user()

    async def test_get_user_5xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_user method of ThermoworksCloud when a 5xx error is returned.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_user method raises an exception when a 5xx error is returned.
        """
        # Setup
        core_test_object.expect_get_user(access_token=TEST_ID_TOKEN).respond_with_data(
            status=500, response_data=b"Internal error"
        )
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_user()

    async def test_get_user_exception_throws(self, auth: Auth):
        """This test checks if the get_user method raises an exception when an exceptions is
        thrown while processing the request.
        """
        # Setup
        with patch(
            "aiohttp.ClientSession.request", new_callable=AsyncMock
        ) as mock_function:
            mock_function.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            try:
                await thermoworks_cloud.get_user()
                assert False, "Should have raised an exception"
            except RuntimeError as e:
                assert e is not None, "Errors should not be swallowed"

    async def test_get_user_read_error_response_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_user method of ThermoworksCloud when an error is returned and the response
        cannot be read.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_user method raises an exception is thrown while readin the
        body of the response
        """
        # Setup
        core_test_object.expect_get_user(access_token=TEST_ID_TOKEN).respond_with_data(
            status=400, response_data=b"Bad Request"
        )
        thermoworks_cloud = ThermoworksCloud(auth)
        with patch("aiohttp.ClientResponse.text", new_callable=AsyncMock) as get_text:
            get_text.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            with pytest.raises(RuntimeError):
                await thermoworks_cloud.get_user()

    async def test_get_device(self, auth: Auth, core_test_object: CoreTestObject):
        """Test the get_device method of ThermoworksCloud.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_device method returns the expected Device object when a valid
        device serial number is provided.
        """
        # Setup
        core_test_object.expect_get_device(
            access_token=TEST_ID_TOKEN, device_serial=TEST_DEVICE_ID_0
        ).respond_with_json(GET_DEVICE_RESPONSE)
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        device = await thermoworks_cloud.get_device(TEST_DEVICE_ID_0)

        # Assert
        # Test that each GET_DEVICE_RESPONSE property matches the Device object property
        assert device is not None
        assert device.battery_state == get_field_value(
            GET_DEVICE_RESPONSE, "batteryState"
        )
        assert device.big_query_info.dataset_id == get_field_value(
            GET_DEVICE_RESPONSE, "bigQuery.mapValue.fields.datasetId"
        )
        assert device.big_query_info.table_id == get_field_value(
            GET_DEVICE_RESPONSE, "bigQuery.mapValue.fields.tableId"
        )
        assert device.battery == int(
            get_field_value(GET_DEVICE_RESPONSE, "battery"))
        assert iso_instant(device.last_archive) == get_field_value(
            GET_DEVICE_RESPONSE, "lastArchive"
        )
        assert device.status == get_field_value(GET_DEVICE_RESPONSE, "status")
        assert iso_instant(device.last_wifi_connection) == get_field_value(
            GET_DEVICE_RESPONSE, "lastWifiConnection"
        )
        assert iso_instant(device.last_bluetooth_connection) == get_field_value(
            GET_DEVICE_RESPONSE, "lastBluetoothConnection"
        )
        assert device.type == get_field_value(GET_DEVICE_RESPONSE, "type")
        assert iso_instant(device.last_telemetry_saved) == get_field_value(
            GET_DEVICE_RESPONSE, "lastTelemetrySaved"
        )
        assert device.pending_load == get_field_value(
            GET_DEVICE_RESPONSE, "pendingLoad"
        )
        assert device.export_version == get_field_value(
            GET_DEVICE_RESPONSE, "exportVersion"
        )
        assert iso_instant(device.session_start) == get_field_value(
            GET_DEVICE_RESPONSE, "sessionStart"
        )
        assert device.device_id == get_field_value(
            GET_DEVICE_RESPONSE, "deviceId")
        assert device.serial == get_field_value(GET_DEVICE_RESPONSE, "serial")
        assert device.transmit_interval_in_seconds == int(
            get_field_value(GET_DEVICE_RESPONSE, "transmitIntervalInSeconds")
        )
        assert device.device_display_units == get_field_value(
            GET_DEVICE_RESPONSE, "deviceDisplayUnits"
        )
        assert device.iot_device_id == get_field_value(
            GET_DEVICE_RESPONSE, "iotDeviceId"
        )
        assert device.label == get_field_value(GET_DEVICE_RESPONSE, "label")
        assert iso_instant(device.last_purged) == get_field_value(
            GET_DEVICE_RESPONSE, "lastPurged"
        )
        assert device.battery_alert_sent == get_field_value(
            GET_DEVICE_RESPONSE, "batteryAlertSent"
        )
        assert device.color == get_field_value(GET_DEVICE_RESPONSE, "color")
        assert device.firmware == get_field_value(
            GET_DEVICE_RESPONSE, "firmware")
        assert device.thumbnail == get_field_value(
            GET_DEVICE_RESPONSE, "thumbnail")
        assert device.wifi_strength == int(
            get_field_value(GET_DEVICE_RESPONSE, "wifi_stength")
        )
        assert device.recording_interval_in_seconds == int(
            get_field_value(GET_DEVICE_RESPONSE, "recordingIntervalInSeconds")
        )
        assert device.account_id == get_field_value(
            GET_DEVICE_RESPONSE, "accountId")
        assert iso_instant(device.last_seen) == get_field_value(
            GET_DEVICE_RESPONSE, "lastSeen"
        )
        assert device.device_name == get_field_value(
            GET_DEVICE_RESPONSE, "device")

    async def test_get_device_4xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_device method of ThermoworksCloud when a 4xx error is returned.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_device method raises an exception when a 4xx error is returned.
        """
        # Setup
        core_test_object.expect_get_device(
            access_token=TEST_ID_TOKEN, device_serial=TEST_DEVICE_ID_0
        ).respond_with_data(status=400, response_data=b"Bad Request")
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_device(TEST_DEVICE_ID_0)

    async def test_get_device_5xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_device method of ThermoworksCloud when a 5xx error is returned.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_device method raises an exception when a 4xx error is returned.
        """
        # Setup
        core_test_object.expect_get_device(
            access_token=TEST_ID_TOKEN, device_serial=TEST_DEVICE_ID_0
        ).respond_with_data(status=500, response_data=b"Internal error")
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_device(TEST_DEVICE_ID_0)

    async def test_get_device_exception_throws(self, auth: Auth):
        """Test the get_device method of ThermoworksCloud when an exception is thrown
        Args:
            auth (Auth): A mock Auth object.

        This test checks if the get_device method raises an exception when an exception is thrown
        during an http request
        """
        # Setup
        with patch(
            "aiohttp.ClientSession.request", new_callable=AsyncMock
        ) as mock_request:
            mock_request.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            try:
                await thermoworks_cloud.get_device(TEST_DEVICE_ID_0)
                assert False, "Should have thrown an exception"
            except RuntimeError as e:
                assert e is not None, "Errors should not be swallowed"

    async def test_get_device_read_error_response_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """This test checks if the get_device method raises an exception when an exception is thrown
        while reading the body of the response
        """
        # Setup

        core_test_object.expect_get_device(
            access_token=TEST_ID_TOKEN, device_serial=TEST_DEVICE_ID_0
        ).respond_with_data(status=500, response_data="Internal error")
        with patch("aiohttp.ClientResponse.text", new_callable=AsyncMock) as get_text:
            get_text.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            with pytest.raises(RuntimeError):
                await thermoworks_cloud.get_device(TEST_DEVICE_ID_0)

    async def test_get_device_channel(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """This test checks if the get_device_channel method returns the expected DeviceChannel 
        object when valid device serial and channel numbers are provided.
        """
        # Setup
        test_device_serial = "test_device_serial"
        test_device_channel = "test_device_channel"
        core_test_object.expect_get_device_channel(
            access_token=TEST_ID_TOKEN,
            device_serial=test_device_serial,
            channel=test_device_channel,
        ).respond_with_json(GET_DEVICE_CHANNEL_RESPONSE)
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        channel = await thermoworks_cloud.get_device_channel(
            test_device_serial, test_device_channel
        )

        # Assert
        assert channel is not None
        assert iso_instant(channel.last_telemetry_saved) == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "lastTelemetrySaved"
        )
        assert channel.value == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "value")

        assert channel.minimum is not None
        assert iso_instant(channel.minimum.date_reading) == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "minimum.mapValue.fields.dateReading"
        )
        assert channel.minimum.reading.value == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE,
            "minimum.mapValue.fields.reading.mapValue.fields.value",
        )
        assert channel.minimum.reading.units == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE,
            "minimum.mapValue.fields.reading.mapValue.fields.units",
        )

        assert channel.units == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "units")
        assert channel.status == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "status")
        assert channel.type == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "type")
        assert channel.label == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "label")
        assert iso_instant(channel.last_seen) == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "lastSeen"
        )

        assert channel.alarm_high is not None
        assert channel.alarm_high.enabled == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "alarmHigh.mapValue.fields.enabled"
        )
        assert channel.alarm_high.units == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "alarmHigh.mapValue.fields.units"
        )
        assert channel.alarm_high.alarming == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "alarmHigh.mapValue.fields.alarming"
        )
        assert channel.alarm_high.value == int(
            get_field_value(
                GET_DEVICE_CHANNEL_RESPONSE, "alarmHigh.mapValue.fields.value"
            )
        )

        assert channel.number == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "number")
        assert channel.maximum is not None
        assert channel.maximum.reading.value == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE,
            "maximum.mapValue.fields.reading.mapValue.fields.value",
        )
        assert channel.maximum.reading.units == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE,
            "maximum.mapValue.fields.reading.mapValue.fields.units",
        )

        assert channel.alarm_low is not None
        assert channel.alarm_low.enabled == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "alarmLow.mapValue.fields.enabled"
        )
        assert channel.alarm_low.units == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "alarmLow.mapValue.fields.units"
        )
        assert channel.alarm_low.alarming == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "alarmLow.mapValue.fields.alarming"
        )
        assert channel.alarm_low.value == int(
            get_field_value(
                GET_DEVICE_CHANNEL_RESPONSE, "alarmLow.mapValue.fields.value"
            )
        )

        assert channel.show_avg_temp == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE, "showAvgTemp"
        )
    async def test_get_device_channel_int_value(
            self, auth: Auth, core_test_object: CoreTestObject
    ):
        """This test checks the value response when the type goes from a float, to an int
        """
        # Setup
        test_device_serial = "test_device_serial"
        test_device_channel = "test_device_channel"
        core_test_object.expect_get_device_channel(
            access_token=TEST_ID_TOKEN,
            device_serial=test_device_serial,
            channel=test_device_channel,
        ).respond_with_json(GET_DEVICE_CHANNEL_RESPONSE_INT)
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        channel = await thermoworks_cloud.get_device_channel(
            test_device_serial, test_device_channel
        )
        # Assert
        assert channel is not None
        assert channel.value == get_field_value(
            GET_DEVICE_CHANNEL_RESPONSE_INT, "value")


    async def test_get_device_channel_4xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_device_channel method of ThermoworksCloud when a 4xx error is returned.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_device_channel method raises an exception when a 4xx error
        is returned.
        """
        # Setup
        test_device_serial = "test_device_serial"
        test_device_channel = "test_device_channel"
        core_test_object.expect_get_device_channel(
            access_token=TEST_ID_TOKEN,
            device_serial=test_device_serial,
            channel=test_device_channel,
        ).respond_with_data(status=400, response_data=b"Bad Request")
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_device_channel(
                test_device_serial, test_device_channel
            )

    async def test_get_device_channel_5xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_device_channel method of ThermoworksCloud when a 5xx error is returned.

        Args:
            auth (Auth): A mock Auth object.
            core_test_object (CoreTestObject): A CoreTestObject instance.

        This test checks if the get_device_channel method raises an exception when a 5xx error
        is returned.
        """
        # Setup
        test_device_serial = "test_device_serial"
        test_device_channel = "test_device_channel"
        core_test_object.expect_get_device_channel(
            access_token=TEST_ID_TOKEN,
            device_serial=test_device_serial,
            channel=test_device_channel,
        ).respond_with_data(status=500, response_data=b"Internal error")
        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_device_channel(
                test_device_serial, test_device_channel
            )

    async def test_get_device_channel_exception_throws(self, auth: Auth):
        """Test the get_device_channel method of ThermoworksCloud when an exception is thrown
        Args:
            auth (Auth): A mock Auth object.

        This test checks if the get_device_channel method raises an exception when an exception is
        thrown during an http request
        """
        # Setup
        with patch(
            "aiohttp.ClientSession.request", new_callable=AsyncMock
        ) as mock_function:
            mock_function.side_effect = RuntimeError("Something went wrong")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            try:
                await thermoworks_cloud.get_device_channel(
                    "test_device_serial", "test_device_channel"
                )
                assert False, "Should have thrown an exception"
            except RuntimeError as e:
                assert e is not None, "Errors should not be swallowed"

    async def test_get_device_channel_read_error_response_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """This test checks if the get_device_channel method raises an exception when an exception
        is thrown while reading the body of the response
        """
        # Setup
        test_device_serial = "test_device_serial"
        test_device_channel = "test_device_channel"
        core_test_object.expect_get_device_channel(
            access_token=TEST_ID_TOKEN,
            device_serial=test_device_serial,
            channel=test_device_channel,
        ).respond_with_data(status=500, response_data="Internal error")
        with patch("aiohttp.ClientResponse.json", new_callable=AsyncMock) as get_json:
            get_json.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            with pytest.raises(RuntimeError):
                await thermoworks_cloud.get_device_channel(
                    test_device_serial, test_device_channel
                )

    # pylint: disable=duplicate-code
    async def test_get_devices(self, auth: Auth, core_test_object: CoreTestObject):
        """Test the get_devices method of ThermoworksCloud.

        This test checks if the get_devices method returns the expected list of Device objects
        when a valid account_id is provided.
        """
        # Setup
        expected_query = {
            "structuredQuery": {
                "from": [{"collectionId": "devices"}],
                "where": {
                    "fieldFilter": {
                        "field": {"fieldPath": "accountId"},
                        "op": "EQUAL",
                        "value": {"stringValue": TEST_ACCOUNT_ID}
                    }
                },
                "orderBy": [{"field": {"fieldPath": "__name__"}, "direction": "ASCENDING"}]
            }
        }

        core_test_object.expect_run_query(
            access_token=TEST_ID_TOKEN, query_body=expected_query
        ).respond_with_json(GET_DEVICES_RESPONSE)

        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        devices = await thermoworks_cloud.get_devices(TEST_ACCOUNT_ID)

        # Assert
        assert devices is not None
        assert len(devices) == 2

        # Check first device (full device)
        device1 = devices[0]
        assert device1.device_id == TEST_DEVICE_ID_0
        assert device1.serial == TEST_DEVICE_ID_0
        assert device1.type == "datalogger"
        assert device1.label == "NODE"
        assert device1.device_name == "node"
        assert device1.account_id == TEST_ACCOUNT_ID
        assert device1.battery == 100
        assert device1.battery_state == "discharging"
        assert device1.firmware == "1.0.26-26"
        assert device1.color == "3f90ca"
        assert device1.wifi_strength == -72
        assert device1.recording_interval_in_seconds == 600
        assert device1.transmit_interval_in_seconds == 7200
        assert device1.iot_device_id == "T123456"
        assert device1.device_display_units == "F"
        assert device1.big_query_info is not None
        assert device1.big_query_info.table_id == TEST_DEVICE_ID_0
        assert device1.big_query_info.dataset_id == "test-dataset-id"

        # Check second device (minimal device)
        device2 = devices[1]
        assert device2.device_id == TEST_DEVICE_ID_1
        assert device2.serial == TEST_DEVICE_ID_1
        assert device2.type == "signals"
        assert device2.label == "SIGNALS"
        assert device2.device_name == "signals"
        assert device2.account_id == TEST_ACCOUNT_ID
        assert device2.status == "NORMAL"

        # Verify that optional fields are handled correctly
        assert device1.big_query_info is not None
        assert device2.big_query_info is None

    # pylint: disable=duplicate-code
    async def test_get_devices_4xx_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_devices method of ThermoworksCloud when a 4xx error is returned."""
        # Setup
        expected_query = {
            "structuredQuery": {
                "from": [{"collectionId": "devices"}],
                "where": {
                    "fieldFilter": {
                        "field": {"fieldPath": "accountId"},
                        "op": "EQUAL",
                        "value": {"stringValue": TEST_ACCOUNT_ID}
                    }
                },
                "orderBy": [{"field": {"fieldPath": "__name__"}, "direction": "ASCENDING"}]
            }
        }

        core_test_object.expect_run_query(
            access_token=TEST_ID_TOKEN, query_body=expected_query
        ).respond_with_data(status=400, response_data=b"Bad Request")

        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        with pytest.raises(RuntimeError):
            await thermoworks_cloud.get_devices(TEST_ACCOUNT_ID)

    async def test_get_devices_empty_response(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_devices method of ThermoworksCloud when an empty response is returned."""
        # Setup
        expected_query = {
            "structuredQuery": {
                "from": [{"collectionId": "devices"}],
                "where": {
                    "fieldFilter": {
                        "field": {"fieldPath": "accountId"},
                        "op": "EQUAL",
                        "value": {"stringValue": TEST_ACCOUNT_ID}
                    }
                },
                "orderBy": [{"field": {"fieldPath": "__name__"}, "direction": "ASCENDING"}]
            }
        }

        core_test_object.expect_run_query(
            access_token=TEST_ID_TOKEN, query_body=expected_query
        ).respond_with_json([])

        thermoworks_cloud = ThermoworksCloud(auth)

        # Act
        devices = await thermoworks_cloud.get_devices(TEST_ACCOUNT_ID)

        # Assert
        assert devices is not None
        assert len(devices) == 0

    async def test_get_devices_read_error_response_throws(
        self, auth: Auth, core_test_object: CoreTestObject
    ):
        """Test the get_devices method of ThermoworksCloud when an error occurs reading 
        the response.
        """
        # Setup
        expected_query = {
            "structuredQuery": {
                "from": [{"collectionId": "devices"}],
                "where": {
                    "fieldFilter": {
                        "field": {"fieldPath": "accountId"},
                        "op": "EQUAL",
                        "value": {"stringValue": TEST_ACCOUNT_ID}
                    }
                },
                "orderBy": [{"field": {"fieldPath": "__name__"}, "direction": "ASCENDING"}]
            }
        }

        core_test_object.expect_run_query(
            access_token=TEST_ID_TOKEN, query_body=expected_query
        ).respond_with_data(status=500, response_data="Internal error")

        with patch("aiohttp.ClientResponse.text", new_callable=AsyncMock) as get_text:
            get_text.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            with pytest.raises(RuntimeError):
                await thermoworks_cloud.get_devices(TEST_ACCOUNT_ID)

    async def test_get_devices_exception_throws(self, auth: Auth):
        """Test the get_devices method of ThermoworksCloud when an exception is thrown 
        during the request.

        This test checks if the get_devices method raises an exception when an exception is thrown
        during an http request.
        """
        # Setup
        with patch(
            "aiohttp.ClientSession.request", new_callable=AsyncMock
        ) as mock_request:
            mock_request.side_effect = RuntimeError("Simulated error")
            thermoworks_cloud = ThermoworksCloud(auth)

            # Act
            try:
                await thermoworks_cloud.get_devices(TEST_ACCOUNT_ID)
                assert False, "Should have thrown an exception"
            except RuntimeError as e:
                assert e is not None, "Errors should not be swallowed"



"""Classes related to User data"""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, Optional, List

from thermoworks_cloud.utils import parse_datetime, map_firestore_fields, parse_nested_object


@dataclass
class EmailLastEvent:  # pylint: disable=too-many-instance-attributes
    """Contains information about the last email sent to a user."""

    reason: Optional[str] = field(default=None, metadata={
                                  "firestore_type": "stringValue"})
    event: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    email: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    bounce_classification: Optional[str] = field(
        default=None,
        metadata={
            "api_name": "bounce_classification",
            "firestore_type": "stringValue"
        }
    )
    tls: Optional[int] = field(default=None, metadata={
                               "firestore_type": "integerValue", "converter": int})
    timestamp: Optional[int] = field(
        default=None, metadata={"firestore_type": "integerValue", "converter": int})
    smtp_id: Optional[str] = field(
        default=None, metadata={"api_name": "smtp-id", "firestore_type": "stringValue"})
    type: Optional[str] = field(default=None, metadata={
                                "firestore_type": "stringValue"})
    sg_message_id: Optional[str] = field(
        default=None, metadata={"api_name": "sg_message_id", "firestore_type": "stringValue"})
    sg_event_id: Optional[str] = field(
        default=None, metadata={"api_name": "sg_event_id", "firestore_type": "stringValue"})

    # Dictionary to store any additional properties not explicitly defined
    additional_properties: Optional[Dict] = None


@dataclass
class DeviceOrderItem:
    """Contains information about a device's order within the users account."""

    device_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    order: Optional[int] = field(default=None, metadata={
                                 "firestore_type": "integerValue", "converter": int})

    # Dictionary to store any additional properties not explicitly defined
    additional_properties: Optional[Dict] = None


@dataclass
class User:  # pylint: disable=too-many-instance-attributes
    """Contains information about a User.

    All fields are optional as different users may have different properties.
    """

    uid: Optional[str] = field(default=None, metadata={
                               "firestore_type": "stringValue"})
    account_id: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    display_name: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    email: Optional[str] = field(default=None, metadata={
                                 "firestore_type": "stringValue"})
    provider: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    time_zone: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    app_version: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    preferred_units: Optional[str] = field(
        default=None, metadata={"firestore_type": "stringValue"})
    locale: Optional[str] = field(default=None, metadata={
                                  "firestore_type": "stringValue"})
    photo_url: Optional[str] = field(
        default=None, metadata={"api_name": "photoURL", "firestore_type": "stringValue"})
    use_24_time: Optional[bool] = field(
        default=None, metadata={"firestore_type": "booleanValue"})
    roles: Optional[Dict[str, bool]] = field(
        default=None, metadata={"map_of": "booleanValue"})
    account_roles: Optional[Dict[str, bool]] = field(
        default=None, metadata={"map_of": "booleanValue"})
    system: Optional[Dict[str, bool]] = field(
        default=None, metadata={"map_of": "booleanValue"})
    notification_settings: Optional[Dict[str, bool]] = field(
        default=None, metadata={"map_of": "booleanValue"})
    fcm_tokens: Optional[Dict[str, bool]] = field(
        default=None, metadata={"map_of": "booleanValue"})
    device_order: Optional[Dict[str, List[DeviceOrderItem]]] = None
    email_last_event: Optional[EmailLastEvent] = None
    export_version: Optional[float] = field(
        default=None, metadata={"firestore_type": "doubleValue"})
    last_seen_in_app: Optional[None] = field(
        default=None, metadata={"firestore_type": "nullValue"})
    last_login: Optional[datetime] = field(
        default=None,
        metadata={
            "firestore_type": "timestampValue",
            "converter": parse_datetime
        }
    )
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None

    # Dictionary to store any additional properties not explicitly defined
    additional_properties: Optional[Dict] = None


def parse_device_order(data: dict) -> Dict[str, List[DeviceOrderItem]]:
    """Parse deviceOrder into a dictionary of account ID to DeviceOrderItem list."""
    if not data or "fields" not in data:
        return {}

    orders = {}
    fields = data["fields"]

    for account_id, devices in fields.items():
        if "arrayValue" not in devices or "values" not in devices["arrayValue"]:
            continue

        device_items = []
        for device in devices["arrayValue"]["values"]:
            if "mapValue" not in device:
                continue

            item = parse_nested_object(device["mapValue"], DeviceOrderItem)
            if item:
                device_items.append(item)

        if device_items:
            orders[account_id] = device_items

    return orders


def document_to_user(document: dict) -> User:
    """Convert a Firestore Document object into a User object."""
    fields = document.get("fields", {})
    user = map_firestore_fields(fields, User)

    # Handle device order separately (complex nested structure)
    if "deviceOrder" in fields and "mapValue" in fields["deviceOrder"]:
        user.device_order = parse_device_order(
            fields["deviceOrder"]["mapValue"])

    # Handle email last event separately (nested object)
    if "emailLastEvent" in fields and "mapValue" in fields["emailLastEvent"]:
        user.email_last_event = parse_nested_object(
            fields["emailLastEvent"]["mapValue"], EmailLastEvent)

    # Document timestamps
    if "createTime" in document:
        user.create_time = parse_datetime(document["createTime"])
    if "updateTime" in document:
        user.update_time = parse_datetime(document["updateTime"])

    return user



"""Test suite for handling missing fields and additional properties."""

from copy import deepcopy

from thermoworks_cloud.models.device import _document_to_device
from thermoworks_cloud.models.device_channel import _document_to_device_channel
from thermoworks_cloud.models.user import document_to_user
from tests.test_data import (
    GET_DEVICE_RESPONSE,
    GET_DEVICE_CHANNEL_RESPONSE,
    USER_RESPONSE,
)


def test_user_missing_fields():
    """Test that the User model handles missing fields gracefully."""
    # Create a copy of the response with some fields removed
    modified_response = deepcopy(USER_RESPONSE)
    del modified_response["fields"]["displayName"]
    del modified_response["fields"]["photoURL"]
    del modified_response["fields"]["system"]

    # Parse the modified response
    user = document_to_user(modified_response)

    # Check that the missing fields are None
    assert user.display_name is None
    assert user.photo_url is None
    assert user.system is None

    # Check that other fields are still present
    assert user.uid is not None
    assert user.email is not None


def test_user_additional_properties():
    """Test that the User model handles additional properties correctly."""
    # Create a copy of the response with an additional field
    modified_response = deepcopy(USER_RESPONSE)
    modified_response["fields"]["newField"] = {"stringValue": "new value"}
    modified_response["fields"]["anotherNewField"] = {"integerValue": "42"}

    # Parse the modified response
    user = document_to_user(modified_response)

    # Check that the additional properties are stored
    assert user.additional_properties is not None
    assert "newField" in user.additional_properties
    assert user.additional_properties["newField"] == "new value"
    assert "anotherNewField" in user.additional_properties
    assert user.additional_properties["anotherNewField"] == "42"


def test_device_missing_fields():
    """Test that the Device model handles missing fields gracefully."""
    # Create a copy of the response with some fields removed
    modified_response = deepcopy(GET_DEVICE_RESPONSE)
    del modified_response["fields"]["label"]
    del modified_response["fields"]["color"]
    del modified_response["fields"]["bigQuery"]

    # Parse the modified response
    device = _document_to_device(modified_response)

    # Check that the missing fields are None
    assert device.label is None
    assert device.color is None
    assert device.big_query_info is None

    # Check that other fields are still present
    assert device.device_id is not None
    assert device.serial is not None


def test_device_additional_properties():
    """Test that the Device model handles additional properties correctly."""
    # Create a copy of the response with an additional field
    modified_response = deepcopy(GET_DEVICE_RESPONSE)
    modified_response["fields"]["newField"] = {"stringValue": "new value"}
    modified_response["fields"]["anotherNewField"] = {"integerValue": "42"}

    # Parse the modified response
    device = _document_to_device(modified_response)

    # Check that the additional properties are stored
    assert device.additional_properties is not None
    assert "newField" in device.additional_properties
    assert device.additional_properties["newField"] == "new value"
    assert "anotherNewField" in device.additional_properties
    assert device.additional_properties["anotherNewField"] == "42"


def test_device_channel_missing_fields():
    """Test that the DeviceChannel model handles missing fields gracefully."""
    # Create a copy of the response with some fields removed
    modified_response = deepcopy(GET_DEVICE_CHANNEL_RESPONSE)
    del modified_response["fields"]["label"]
    del modified_response["fields"]["alarmHigh"]
    del modified_response["fields"]["minimum"]

    # Parse the modified response
    channel = _document_to_device_channel(modified_response)

    # Check that the missing fields are None
    assert channel.label is None
    assert channel.alarm_high is None
    assert channel.minimum is None

    # Check that other fields are still present
    assert channel.value is not None
    assert channel.units is not None


def test_device_channel_additional_properties():
    """Test that the DeviceChannel model handles additional properties correctly."""
    # Create a copy of the response with an additional field
    modified_response = deepcopy(GET_DEVICE_CHANNEL_RESPONSE)
    modified_response["fields"]["newField"] = {"stringValue": "new value"}
    modified_response["fields"]["anotherNewField"] = {"integerValue": "42"}

    # Parse the modified response
    channel = _document_to_device_channel(modified_response)

    # Check that the additional properties are stored
    assert channel.additional_properties is not None
    assert "newField" in channel.additional_properties
    assert channel.additional_properties["newField"] == "new value"
    assert "anotherNewField" in channel.additional_properties
    assert channel.additional_properties["anotherNewField"] == "42"



"""Demonstrates how to use this library to get the devices for the logged in user.

This example shows how to:
1. Get all devices for a user
2. Get all channels for each device
3. Print detailed information about devices and their properties
4. Identify device-specific properties that may not be present in all device types
"""
import asyncio
import json
import os
import pprint
from dataclasses import asdict
from typing import Dict

from aiohttp import ClientSession

from thermoworks_cloud import AuthFactory, ThermoworksCloud, ResourceNotFoundError
from thermoworks_cloud.models.device import Device
from thermoworks_cloud.models.device_channel import DeviceChannel

# Make sure these are defined
email = os.environ["THERMOWORKS_EMAIL"]
password = os.environ["THERMOWORKS_PASSWORD"]


def print_device_info(device: Device, device_channels: list[DeviceChannel]):
    """Print detailed information about a device and its channels."""
    device_dict = asdict(device)

    print(f"\n{'=' * 50}")
    print(f"DEVICE TYPE: {device.type or 'Unknown'}")
    print(f"DEVICE NAME: {device.label or 'Unnamed'}")
    print(f"SERIAL: {device.serial or 'Unknown'}")
    print(f"{'=' * 50}")

    # Print all device properties
    print("\nDEVICE PROPERTIES:")
    pprint.pprint(device_dict, compact=False)

    # Print channel information
    if device_channels:
        print(f"\nCHANNELS ({len(device_channels)}):")
        for i, channel in enumerate(device_channels):
            channel_dict = asdict(channel)
            print(f"\n  Channel {i+1}:")
            pprint.pprint(channel_dict)
    else:
        print("\nNo channels found for this device.")


async def __main__():
    # Use a context manager when providing the session to the auth factory
    async with ClientSession() as session:
        auth = await AuthFactory(session).build_auth(email, password)
        thermoworks = ThermoworksCloud(auth)
        user = await thermoworks.get_user()
        if not user.account_id:
            raise RuntimeError("No account ID found for user")

        devices = await thermoworks.get_devices(user.account_id)
        device_channels_by_device: dict[str, list[DeviceChannel]] = {}

        # Iterate over the device serials and fetch the device document for each
        for device in devices:
            if not device.serial:
                print(
                    f"Cannot get channels for device {device.device_id}, device has no serial")
                continue

            try:
                device_channels = []

                # According to reverse engineering, channels seem to be 1 indexed
                for channel in range(1, 10):
                    try:
                        device_channels.append(
                            await thermoworks.get_device_channel(
                                device_serial=device.serial, channel=str(
                                    channel)
                            )
                        )
                    except ResourceNotFoundError:
                        # Go until there are no more
                        break
                    except Exception as e:
                        print(
                            f"Error getting channel {channel} for device {device.serial}: {e}")
                        continue

                device_channels_by_device[device.serial] = device_channels
            except Exception as e:
                print(f"Error getting device {device.serial}: {e}")
                continue

        # Print information about each device
        print(f"\nFound {len(devices)} devices")

        # Group devices by type
        devices_by_type: Dict[str, list[Device]] = {}
        for device in devices:
            device_type = device.type or "unknown"
            if device_type not in devices_by_type:
                devices_by_type[device_type] = []
            devices_by_type[device_type].append(device)

        # Print summary of device types
        print("\nDEVICE TYPES SUMMARY:")
        for device_type, type_devices in devices_by_type.items():
            print(f"  {device_type}: {len(type_devices)} device(s)")

        # Print detailed information for each device
        for device in devices:
            assert device.serial is not None
            device_channels = device_channels_by_device.get(device.serial, [])
            print_device_info(device, device_channels)

        # Save the data to a JSON file for further analysis
        output_data = {
            "devices": [asdict(device) for device in devices],
            "device_channels": {
                serial: [asdict(channel) for channel in channels]
                for serial, channels in device_channels_by_device.items()
            }
        }

        with open("thermoworks_devices_data.json", "w", encoding="utf-8") as f:
            json.dump(output_data, f, indent=2, default=str)

        print("Data saved to thermoworks_devices_data.json for further analysis")


asyncio.run(__main__())



"""Provides capabilities for mocking authentication related API calls."""

from pytest_httpserver import HTTPServer, RequestHandler

from tests.test_data import (
    API_KEY,
    FIREBASE_APPLICATION_BASE_PATH,
    LOGIN_PATH,
    TOKEN_PATH,
    WEB_CONFIG_PATH,
)


class AuthTestObject:
    """A test object that can be used to mock authentication related API calls."""

    def __init__(self, httpserver: HTTPServer) -> None:
        """Create a new AuthTestObject

        Args:
            httpserver (HTTPServer): The httpserver that the test object will use to mock responses
        """
        self.httpserver = httpserver

    def expect_config(self) -> RequestHandler:
        """Create a request handler for a `get web config` request."""
        headers = {"x-goog-api-key": API_KEY, "accept": "application/json"}
        return self.httpserver.expect_request(WEB_CONFIG_PATH, headers=headers)

    def expect_login(self, email: str, password: str) -> RequestHandler:
        """Create a request handler for a `sign in with email` request."""
        headers = {"Content-Type": "application/json"}
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            LOGIN_PATH,
            headers=headers,
            query_string=query_string,
            json={
                "email": email,
                "password": password,
                "returnSecureToken": True,
            },
        )

    def expect_token_refresh(self, refresh_token: str) -> RequestHandler:
        """Create a request handler for a `refresh token` request."""
        headers = {"Content-Type": "application/json"}
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            TOKEN_PATH,
            headers=headers,
            query_string=query_string,
            json={
                "grant_type": "refresh_token",
                "refresh_token": refresh_token,
            },
        )

    def expect_request(self, path: str, headers: dict) -> RequestHandler:
        """Create a request handler for a generic request to this firestore project."""
        query_string = {"key": API_KEY}
        return self.httpserver.expect_request(
            f"{FIREBASE_APPLICATION_BASE_PATH}/{path}",
            headers=headers,
            query_string=query_string,
        )



"""Model for User Credentials for Firestore."""

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TypedDict


class _UserLoginResponse(TypedDict):
    """See:
    <https://firebase.google.com/docs/reference/rest/auth#section-sign-in-email-password>.
    """

    kind: str
    localId: str
    email: str
    displayName: str
    idToken: str
    registered: bool
    refreshToken: str
    expiresIn: str


class _RefreshTokenResponse(TypedDict):
    """See:
    <https://firebase.google.com/docs/reference/rest/auth#section-refresh-token>.
    """

    access_token: str
    expires_in: str
    token_type: str
    refresh_token: str
    id_token: str
    user_id: str
    project_id: str


def _get_expiration_time(seconds_until_expiration: int):
    """Calculate the exact expiration time of a key.

    Args:
        seconds_until_expiration (int): Number of seconds until the key expires.

    Returns:
        datetime: The expiration time.

    """
    # Get the current time
    current_time = datetime.now()

    # Calculate the expiration time
    return current_time + timedelta(seconds=seconds_until_expiration)


@dataclass
class _UserCredentials:
    """Internal representation of credentials for a user."""

    user_id: str
    access_token: str
    refresh_token: str
    expiration_time: datetime

    @staticmethod
    def from_user_login_response(response: _UserLoginResponse) -> "_UserCredentials":
        """Create UserCredentials from the login API response."""
        return _UserCredentials(
            user_id=response["localId"],
            access_token=response["idToken"],
            refresh_token=response["refreshToken"],
            expiration_time=_get_expiration_time(int(response["expiresIn"])),
        )

    @staticmethod
    def from_refresh_token_response(
        response: _RefreshTokenResponse,
    ) -> "_UserCredentials":
        """Create UserCredentials from the refresh token API response."""
        return _UserCredentials(
            user_id=response["user_id"],
            access_token=response["access_token"],
            refresh_token=response["refresh_token"],
            expiration_time=_get_expiration_time(int(response["expires_in"])),
        )



"""ThermoWorks Cloud data models."""

from .device import Device, BigQueryInfo
from .device_channel import Alarm, DeviceChannel, MinMaxReading, Reading
from .user import DeviceOrderItem, EmailLastEvent, User


__all__ = ["Device", "BigQueryInfo", "Alarm",
           "DeviceChannel", "MinMaxReading", "Reading", "DeviceOrderItem", "EmailLastEvent", "User"]
